from pathlib import Path

try:
    from tqdm.auto import tqdm
except ImportError:

    def tqdm(iterable, *args, **kwargs):
        return iterable


BASE_URL = "https://public.ukp.informatik.tu-darmstadt.de/thakur/BEIR/datasets/{}.zip"


def clean_results_keys(beir_results):
    return {k.split("@")[-1]: v for k, v in beir_results.items()}


def postprocess_results_for_eval(results, scores, query_ids):
    """
    Given the queried results and scores output by BM25S, postprocess them
    to be compatible with BEIR evaluation functions.
    query_ids is a list of query ids in the same order as the results.
    """

    results_record = [
        {"id": qid, "hits": results[i], "scores": list(scores[i])}
        for i, qid in enumerate(query_ids)
    ]

    result_dict_for_eval = {
        res["id"]: {
            docid: float(score) for docid, score in zip(res["hits"], res["scores"])
        }
        for res in results_record
    }

    return result_dict_for_eval


def merge_cqa_dupstack(data_path):
    try:
        import ujson
    except ImportError:
        import json as ujson

    data_path = Path(data_path)
    dataset = data_path.name
    assert dataset == "cqadupstack", "Dataset must be CQADupStack"

    # check if corpus.jsonl exists
    corpus_path = data_path / "corpus.jsonl"
    if not corpus_path.exists():
        # combine all the corpus files into one
        # corpus files are located under cqadupstack/<name>/corpus.jsonl
        corpus_files = list(data_path.glob("*/corpus.jsonl"))
        with open(corpus_path, "w") as f:
            for file in tqdm(corpus_files, desc="Merging Corpus", leave=False):
                # get the name of the corpus
                corpus_name = file.parent.name

                with open(file, "r") as f2:
                    for line in tqdm(
                        f2, desc=f"Merging {corpus_name} Corpus", leave=False
                    ):
                        # first, read with ujson
                        line = ujson.loads(line)
                        # add the corpus name to _id
                        line["_id"] = f"{corpus_name}_{line['_id']}"
                        # write back to file
                        f.write(ujson.dumps(line))
                        f.write("\n")

    # now, do the same for queries.jsonl
    queries_path = data_path / "queries.jsonl"
    if not queries_path.exists():
        queries_files = list(data_path.glob("*/queries.jsonl"))
        with open(queries_path, "w") as f:
            for file in tqdm(queries_files, desc="Merging Queries", leave=False):
                # get the name of the corpus
                corpus_name = file.parent.name

                with open(file, "r") as f2:
                    for line in tqdm(
                        f2, desc=f"Merging {corpus_name} Queries", leave=False
                    ):
                        # first, read with ujson
                        line = ujson.loads(line)
                        # add the corpus name to _id
                        line["_id"] = f"{corpus_name}_{line['_id']}"
                        # write back to file
                        f.write(ujson.dumps(line))
                        f.write("\n")

    # now, do the same for qrels/test.tsv
    qrels_path = data_path / "qrels" / "test.tsv"
    qrels_path.parent.mkdir(parents=True, exist_ok=True)

    if not qrels_path.exists():
        qrels_files = list(data_path.glob("*/qrels/test.tsv"))
        with open(qrels_path, "w") as f:
            # First, write the columns: query-id	corpus-id	score
            f.write("query-id\tcorpus-id\tscore\n")
            for file in tqdm(qrels_files, desc="Merging Qrels", leave=False):
                # get the name of the corpus
                corpus_name = file.parent.parent.name
                with open(file, "r") as f2:
                    # skip first line
                    next(f2)

                    for line in tqdm(
                        f2, desc=f"Merging {corpus_name} Qrels", leave=False
                    ):
                        # since it's a tsv, split by tab
                        qid, cid, score = line.strip().split("\t")
                        # add the corpus name to _id
                        qid = f"{corpus_name}_{qid}"
                        cid = f"{corpus_name}_{cid}"
                        # write back to file
                        f.write(f"{qid}\t{cid}\t{score}\n")



STOPWORDS_EN = (
    "a",
    "an",
    "and",
    "are",
    "as",
    "at",
    "be",
    "but",
    "by",
    "for",
    "if",
    "in",
    "into",
    "is",
    "it",
    "no",
    "not",
    "of",
    "on",
    "or",
    "such",
    "that",
    "the",
    "their",
    "then",
    "there",
    "these",
    "they",
    "this",
    "to",
    "was",
    "will",
    "with",
)

STOPWORDS_EN_PLUS = (
    "a",
    "about",
    "above",
    "after",
    "again",
    "against",
    "ain",
    "all",
    "am",
    "an",
    "and",
    "any",
    "are",
    "aren",
    "aren't",
    "as",
    "at",
    "be",
    "because",
    "been",
    "before",
    "being",
    "below",
    "between",
    "both",
    "but",
    "by",
    "can",
    "couldn",
    "couldn't",
    "d",
    "did",
    "didn",
    "didn't",
    "do",
    "does",
    "doesn",
    "doesn't",
    "doing",
    "don",
    "don't",
    "down",
    "during",
    "each",
    "few",
    "for",
    "from",
    "further",
    "had",
    "hadn",
    "hadn't",
    "has",
    "hasn",
    "hasn't",
    "have",
    "haven",
    "haven't",
    "having",
    "he",
    "her",
    "here",
    "hers",
    "herself",
    "him",
    "himself",
    "his",
    "how",
    "i",
    "if",
    "in",
    "into",
    "is",
    "isn",
    "isn't",
    "it",
    "it's",
    "its",
    "itself",
    "just",
    "ll",
    "m",
    "ma",
    "me",
    "mightn",
    "mightn't",
    "more",
    "most",
    "mustn",
    "mustn't",
    "my",
    "myself",
    "needn",
    "needn't",
    "no",
    "nor",
    "not",
    "now",
    "o",
    "of",
    "off",
    "on",
    "once",
    "only",
    "or",
    "other",
    "our",
    "ours",
    "ourselves",
    "out",
    "over",
    "own",
    "re",
    "s",
    "same",
    "shan",
    "shan't",
    "she",
    "she's",
    "should",
    "should've",
    "shouldn",
    "shouldn't",
    "so",
    "some",
    "such",
    "t",
    "than",
    "that",
    "that'll",
    "the",
    "their",
    "theirs",
    "them",
    "themselves",
    "then",
    "there",
    "these",
    "they",
    "this",
    "those",
    "through",
    "to",
    "too",
    "under",
    "until",
    "up",
    "ve",
    "very",
    "was",
    "wasn",
    "wasn't",
    "we",
    "were",
    "weren",
    "weren't",
    "what",
    "when",
    "where",
    "which",
    "while",
    "who",
    "whom",
    "why",
    "will",
    "with",
    "won",
    "won't",
    "wouldn",
    "wouldn't",
    "y",
    "you",
    "you'd",
    "you'll",
    "you're",
    "you've",
    "your",
    "yours",
    "yourself",
    "yourselves",
)

STOPWORDS_GERMAN = (
    "aber",
    "alle",
    "allem",
    "allen",
    "aller",
    "alles",
    "als",
    "also",
    "am",
    "an",
    "ander",
    "andere",
    "anderem",
    "anderen",
    "anderer",
    "anderes",
    "anderm",
    "andern",
    "anderr",
    "anders",
    "auch",
    "auf",
    "aus",
    "bei",
    "bin",
    "bis",
    "bist",
    "da",
    "damit",
    "dann",
    "das",
    "dass",
    "dasselbe",
    "dazu",
    "daß",
    "dein",
    "deine",
    "deinem",
    "deinen",
    "deiner",
    "deines",
    "dem",
    "demselben",
    "den",
    "denn",
    "denselben",
    "der",
    "derer",
    "derselbe",
    "derselben",
    "des",
    "desselben",
    "dessen",
    "dich",
    "die",
    "dies",
    "diese",
    "dieselbe",
    "dieselben",
    "diesem",
    "diesen",
    "dieser",
    "dieses",
    "dir",
    "doch",
    "dort",
    "du",
    "durch",
    "ein",
    "eine",
    "einem",
    "einen",
    "einer",
    "eines",
    "einig",
    "einige",
    "einigem",
    "einigen",
    "einiger",
    "einiges",
    "einmal",
    "er",
    "es",
    "etwas",
    "euch",
    "euer",
    "eure",
    "eurem",
    "euren",
    "eurer",
    "eures",
    "für",
    "gegen",
    "gewesen",
    "hab",
    "habe",
    "haben",
    "hat",
    "hatte",
    "hatten",
    "hier",
    "hin",
    "hinter",
    "ich",
    "ihm",
    "ihn",
    "ihnen",
    "ihr",
    "ihre",
    "ihrem",
    "ihren",
    "ihrer",
    "ihres",
    "im",
    "in",
    "indem",
    "ins",
    "ist",
    "jede",
    "jedem",
    "jeden",
    "jeder",
    "jedes",
    "jene",
    "jenem",
    "jenen",
    "jener",
    "jenes",
    "jetzt",
    "kann",
    "kein",
    "keine",
    "keinem",
    "keinen",
    "keiner",
    "keines",
    "können",
    "könnte",
    "machen",
    "man",
    "manche",
    "manchem",
    "manchen",
    "mancher",
    "manches",
    "mein",
    "meine",
    "meinem",
    "meinen",
    "meiner",
    "meines",
    "mich",
    "mir",
    "mit",
    "muss",
    "musste",
    "nach",
    "nicht",
    "nichts",
    "noch",
    "nun",
    "nur",
    "ob",
    "oder",
    "ohne",
    "sehr",
    "sein",
    "seine",
    "seinem",
    "seinen",
    "seiner",
    "seines",
    "selbst",
    "sich",
    "sie",
    "sind",
    "so",
    "solche",
    "solchem",
    "solchen",
    "solcher",
    "solches",
    "soll",
    "sollte",
    "sondern",
    "sonst",
    "um",
    "und",
    "uns",
    "unser",
    "unsere",
    "unserem",
    "unseren",
    "unseres",
    "unter",
    "viel",
    "vom",
    "von",
    "vor",
    "war",
    "waren",
    "warst",
    "was",
    "weg",
    "weil",
    "weiter",
    "welche",
    "welchem",
    "welchen",
    "welcher",
    "welches",
    "wenn",
    "werde",
    "werden",
    "wie",
    "wieder",
    "will",
    "wir",
    "wird",
    "wirst",
    "wo",
    "wollen",
    "wollte",
    "während",
    "würde",
    "würden",
    "zu",
    "zum",
    "zur",
    "zwar",
    "zwischen",
    "über",
)
STOPWORDS_DUTCH = (
    "aan",
    "al",
    "alles",
    "als",
    "altijd",
    "andere",
    "ben",
    "bij",
    "daar",
    "dan",
    "dat",
    "de",
    "der",
    "deze",
    "die",
    "dit",
    "doch",
    "doen",
    "door",
    "dus",
    "een",
    "eens",
    "en",
    "er",
    "ge",
    "geen",
    "geweest",
    "haar",
    "had",
    "heb",
    "hebben",
    "heeft",
    "hem",
    "het",
    "hier",
    "hij",
    "hoe",
    "hun",
    "iemand",
    "iets",
    "ik",
    "in",
    "is",
    "ja",
    "je",
    "kan",
    "kon",
    "kunnen",
    "maar",
    "me",
    "meer",
    "men",
    "met",
    "mij",
    "mijn",
    "moet",
    "na",
    "naar",
    "niet",
    "niets",
    "nog",
    "nu",
    "of",
    "om",
    "omdat",
    "onder",
    "ons",
    "ook",
    "op",
    "over",
    "reeds",
    "te",
    "tegen",
    "toch",
    "toen",
    "tot",
    "u",
    "uit",
    "uw",
    "van",
    "veel",
    "voor",
    "want",
    "waren",
    "was",
    "wat",
    "werd",
    "wezen",
    "wie",
    "wil",
    "worden",
    "wordt",
    "zal",
    "ze",
    "zelf",
    "zich",
    "zij",
    "zijn",
    "zo",
    "zonder",
    "zou",
)
STOPWORDS_FRENCH = (
    "ai",
    "aie",
    "aient",
    "aies",
    "ait",
    "as",
    "au",
    "aura",
    "aurai",
    "auraient",
    "aurais",
    "aurait",
    "auras",
    "aurez",
    "auriez",
    "aurions",
    "aurons",
    "auront",
    "aux",
    "avaient",
    "avais",
    "avait",
    "avec",
    "avez",
    "aviez",
    "avions",
    "avons",
    "ayant",
    "ayante",
    "ayantes",
    "ayants",
    "ayez",
    "ayons",
    "c",
    "ce",
    "ces",
    "d",
    "dans",
    "de",
    "des",
    "du",
    "elle",
    "en",
    "es",
    "est",
    "et",
    "eu",
    "eue",
    "eues",
    "eurent",
    "eus",
    "eusse",
    "eussent",
    "eusses",
    "eussiez",
    "eussions",
    "eut",
    "eux",
    "eûmes",
    "eût",
    "eûtes",
    "furent",
    "fus",
    "fusse",
    "fussent",
    "fusses",
    "fussiez",
    "fussions",
    "fut",
    "fûmes",
    "fût",
    "fûtes",
    "il",
    "ils",
    "j",
    "je",
    "l",
    "la",
    "le",
    "les",
    "leur",
    "lui",
    "m",
    "ma",
    "mais",
    "me",
    "mes",
    "moi",
    "mon",
    "même",
    "n",
    "ne",
    "nos",
    "notre",
    "nous",
    "on",
    "ont",
    "ou",
    "par",
    "pas",
    "pour",
    "qu",
    "que",
    "qui",
    "s",
    "sa",
    "se",
    "sera",
    "serai",
    "seraient",
    "serais",
    "serait",
    "seras",
    "serez",
    "seriez",
    "serions",
    "serons",
    "seront",
    "ses",
    "soient",
    "sois",
    "soit",
    "sommes",
    "son",
    "sont",
    "soyez",
    "soyons",
    "suis",
    "sur",
    "t",
    "ta",
    "te",
    "tes",
    "toi",
    "ton",
    "tu",
    "un",
    "une",
    "vos",
    "votre",
    "vous",
    "y",
    "à",
    "étaient",
    "étais",
    "était",
    "étant",
    "étante",
    "étantes",
    "étants",
    "étiez",
    "étions",
    "été",
    "étée",
    "étées",
    "étés",
    "êtes",
)
STOPWORDS_SPANISH = (
    "a",
    "al",
    "algo",
    "algunas",
    "algunos",
    "ante",
    "antes",
    "como",
    "con",
    "contra",
    "cual",
    "cuando",
    "de",
    "del",
    "desde",
    "donde",
    "durante",
    "e",
    "el",
    "ella",
    "ellas",
    "ellos",
    "en",
    "entre",
    "era",
    "erais",
    "eran",
    "eras",
    "eres",
    "es",
    "esa",
    "esas",
    "ese",
    "eso",
    "esos",
    "esta",
    "estaba",
    "estabais",
    "estaban",
    "estabas",
    "estad",
    "estada",
    "estadas",
    "estado",
    "estados",
    "estamos",
    "estando",
    "estar",
    "estaremos",
    "estará",
    "estarán",
    "estarás",
    "estaré",
    "estaréis",
    "estaría",
    "estaríais",
    "estaríamos",
    "estarían",
    "estarías",
    "estas",
    "este",
    "estemos",
    "esto",
    "estos",
    "estoy",
    "estuve",
    "estuviera",
    "estuvierais",
    "estuvieran",
    "estuvieras",
    "estuvieron",
    "estuviese",
    "estuvieseis",
    "estuviesen",
    "estuvieses",
    "estuvimos",
    "estuviste",
    "estuvisteis",
    "estuviéramos",
    "estuviésemos",
    "estuvo",
    "está",
    "estábamos",
    "estáis",
    "están",
    "estás",
    "esté",
    "estéis",
    "estén",
    "estés",
    "fue",
    "fuera",
    "fuerais",
    "fueran",
    "fueras",
    "fueron",
    "fuese",
    "fueseis",
    "fuesen",
    "fueses",
    "fui",
    "fuimos",
    "fuiste",
    "fuisteis",
    "fuéramos",
    "fuésemos",
    "ha",
    "habida",
    "habidas",
    "habido",
    "habidos",
    "habiendo",
    "habremos",
    "habrá",
    "habrán",
    "habrás",
    "habré",
    "habréis",
    "habría",
    "habríais",
    "habríamos",
    "habrían",
    "habrías",
    "habéis",
    "había",
    "habíais",
    "habíamos",
    "habían",
    "habías",
    "han",
    "has",
    "hasta",
    "hay",
    "haya",
    "hayamos",
    "hayan",
    "hayas",
    "hayáis",
    "he",
    "hemos",
    "hube",
    "hubiera",
    "hubierais",
    "hubieran",
    "hubieras",
    "hubieron",
    "hubiese",
    "hubieseis",
    "hubiesen",
    "hubieses",
    "hubimos",
    "hubiste",
    "hubisteis",
    "hubiéramos",
    "hubiésemos",
    "hubo",
    "la",
    "las",
    "le",
    "les",
    "lo",
    "los",
    "me",
    "mi",
    "mis",
    "mucho",
    "muchos",
    "muy",
    "más",
    "mí",
    "mía",
    "mías",
    "mío",
    "míos",
    "nada",
    "ni",
    "no",
    "nos",
    "nosotras",
    "nosotros",
    "nuestra",
    "nuestras",
    "nuestro",
    "nuestros",
    "o",
    "os",
    "otra",
    "otras",
    "otro",
    "otros",
    "para",
    "pero",
    "poco",
    "por",
    "porque",
    "que",
    "quien",
    "quienes",
    "qué",
    "se",
    "sea",
    "seamos",
    "sean",
    "seas",
    "sentid",
    "sentida",
    "sentidas",
    "sentido",
    "sentidos",
    "seremos",
    "será",
    "serán",
    "serás",
    "seré",
    "seréis",
    "sería",
    "seríais",
    "seríamos",
    "serían",
    "serías",
    "seáis",
    "siente",
    "sin",
    "sintiendo",
    "sobre",
    "sois",
    "somos",
    "son",
    "soy",
    "su",
    "sus",
    "suya",
    "suyas",
    "suyo",
    "suyos",
    "sí",
    "también",
    "tanto",
    "te",
    "tendremos",
    "tendrá",
    "tendrán",
    "tendrás",
    "tendré",
    "tendréis",
    "tendría",
    "tendríais",
    "tendríamos",
    "tendrían",
    "tendrías",
    "tened",
    "tenemos",
    "tenga",
    "tengamos",
    "tengan",
    "tengas",
    "tengo",
    "tengáis",
    "tenida",
    "tenidas",
    "tenido",
    "tenidos",
    "teniendo",
    "tenéis",
    "tenía",
    "teníais",
    "teníamos",
    "tenían",
    "tenías",
    "ti",
    "tiene",
    "tienen",
    "tienes",
    "todo",
    "todos",
    "tu",
    "tus",
    "tuve",
    "tuviera",
    "tuvierais",
    "tuvieran",
    "tuvieras",
    "tuvieron",
    "tuviese",
    "tuvieseis",
    "tuviesen",
    "tuvieses",
    "tuvimos",
    "tuviste",
    "tuvisteis",
    "tuviéramos",
    "tuviésemos",
    "tuvo",
    "tuya",
    "tuyas",
    "tuyo",
    "tuyos",
    "tú",
    "un",
    "una",
    "uno",
    "unos",
    "vosotras",
    "vosotros",
    "vuestra",
    "vuestras",
    "vuestro",
    "vuestros",
    "y",
    "ya",
    "yo",
    "él",
    "éramos",
)
STOPWORDS_PORTUGUESE = (
    "a",
    "ao",
    "aos",
    "aquela",
    "aquelas",
    "aquele",
    "aqueles",
    "aquilo",
    "as",
    "até",
    "com",
    "como",
    "da",
    "das",
    "de",
    "dela",
    "delas",
    "dele",
    "deles",
    "depois",
    "do",
    "dos",
    "e",
    "ela",
    "elas",
    "ele",
    "eles",
    "em",
    "entre",
    "era",
    "eram",
    "essa",
    "essas",
    "esse",
    "esses",
    "esta",
    "estamos",
    "estar",
    "estas",
    "estava",
    "estavam",
    "este",
    "esteja",
    "estejam",
    "estejamos",
    "estes",
    "esteve",
    "estive",
    "estivemos",
    "estiver",
    "estivera",
    "estiveram",
    "estiverem",
    "estivermos",
    "estivesse",
    "estivessem",
    "estivéramos",
    "estivéssemos",
    "estou",
    "está",
    "estávamos",
    "estão",
    "eu",
    "foi",
    "fomos",
    "for",
    "fora",
    "foram",
    "forem",
    "formos",
    "fosse",
    "fossem",
    "fui",
    "fôramos",
    "fôssemos",
    "haja",
    "hajam",
    "hajamos",
    "havemos",
    "haver",
    "hei",
    "houve",
    "houvemos",
    "houver",
    "houvera",
    "houveram",
    "houverei",
    "houverem",
    "houveremos",
    "houveria",
    "houveriam",
    "houvermos",
    "houverá",
    "houverão",
    "houveríamos",
    "houvesse",
    "houvessem",
    "houvéramos",
    "houvéssemos",
    "há",
    "hão",
    "isso",
    "isto",
    "já",
    "lhe",
    "lhes",
    "mais",
    "mas",
    "me",
    "mesmo",
    "meu",
    "meus",
    "minha",
    "minhas",
    "muito",
    "na",
    "nas",
    "nem",
    "no",
    "nos",
    "nossa",
    "nossas",
    "nosso",
    "nossos",
    "num",
    "numa",
    "não",
    "nós",
    "o",
    "os",
    "ou",
    "para",
    "pela",
    "pelas",
    "pelo",
    "pelos",
    "por",
    "qual",
    "quando",
    "que",
    "quem",
    "se",
    "seja",
    "sejam",
    "sejamos",
    "sem",
    "ser",
    "serei",
    "seremos",
    "seria",
    "seriam",
    "será",
    "serão",
    "seríamos",
    "seu",
    "seus",
    "somos",
    "sou",
    "sua",
    "suas",
    "são",
    "só",
    "também",
    "te",
    "tem",
    "temos",
    "tenha",
    "tenham",
    "tenhamos",
    "tenho",
    "terei",
    "teremos",
    "teria",
    "teriam",
    "terá",
    "terão",
    "teríamos",
    "teu",
    "teus",
    "teve",
    "tinha",
    "tinham",
    "tive",
    "tivemos",
    "tiver",
    "tivera",
    "tiveram",
    "tiverem",
    "tivermos",
    "tivesse",
    "tivessem",
    "tivéramos",
    "tivéssemos",
    "tu",
    "tua",
    "tuas",
    "tém",
    "tínhamos",
    "um",
    "uma",
    "você",
    "vocês",
    "vos",
    "à",
    "às",
    "é",
    "éramos",
)
STOPWORDS_ITALIAN = (
    "a",
    "abbia",
    "abbiamo",
    "abbiano",
    "abbiate",
    "ad",
    "agl",
    "agli",
    "ai",
    "al",
    "all",
    "alla",
    "alle",
    "allo",
    "anche",
    "avemmo",
    "avendo",
    "avesse",
    "avessero",
    "avessi",
    "avessimo",
    "aveste",
    "avesti",
    "avete",
    "aveva",
    "avevamo",
    "avevano",
    "avevate",
    "avevi",
    "avevo",
    "avrai",
    "avranno",
    "avrebbe",
    "avrebbero",
    "avrei",
    "avremmo",
    "avremo",
    "avreste",
    "avresti",
    "avrete",
    "avrà",
    "avrò",
    "avuta",
    "avute",
    "avuti",
    "avuto",
    "c",
    "che",
    "chi",
    "ci",
    "coi",
    "col",
    "come",
    "con",
    "contro",
    "cui",
    "da",
    "dagl",
    "dagli",
    "dai",
    "dal",
    "dall",
    "dalla",
    "dalle",
    "dallo",
    "degl",
    "degli",
    "dei",
    "del",
    "dell",
    "della",
    "delle",
    "dello",
    "di",
    "dov",
    "dove",
    "e",
    "ebbe",
    "ebbero",
    "ebbi",
    "ed",
    "era",
    "erano",
    "eravamo",
    "eravate",
    "eri",
    "ero",
    "essendo",
    "faccia",
    "facciamo",
    "facciano",
    "facciate",
    "faccio",
    "facemmo",
    "facendo",
    "facesse",
    "facessero",
    "facessi",
    "facessimo",
    "faceste",
    "facesti",
    "faceva",
    "facevamo",
    "facevano",
    "facevate",
    "facevi",
    "facevo",
    "fai",
    "fanno",
    "farai",
    "faranno",
    "farebbe",
    "farebbero",
    "farei",
    "faremmo",
    "faremo",
    "fareste",
    "faresti",
    "farete",
    "farà",
    "farò",
    "fece",
    "fecero",
    "feci",
    "fosse",
    "fossero",
    "fossi",
    "fossimo",
    "foste",
    "fosti",
    "fu",
    "fui",
    "fummo",
    "furono",
    "gli",
    "ha",
    "hai",
    "hanno",
    "ho",
    "i",
    "il",
    "in",
    "io",
    "l",
    "la",
    "le",
    "lei",
    "li",
    "lo",
    "loro",
    "lui",
    "ma",
    "mi",
    "mia",
    "mie",
    "miei",
    "mio",
    "ne",
    "negl",
    "negli",
    "nei",
    "nel",
    "nell",
    "nella",
    "nelle",
    "nello",
    "noi",
    "non",
    "nostra",
    "nostre",
    "nostri",
    "nostro",
    "o",
    "per",
    "perché",
    "più",
    "quale",
    "quanta",
    "quante",
    "quanti",
    "quanto",
    "quella",
    "quelle",
    "quelli",
    "quello",
    "questa",
    "queste",
    "questi",
    "questo",
    "sarai",
    "saranno",
    "sarebbe",
    "sarebbero",
    "sarei",
    "saremmo",
    "saremo",
    "sareste",
    "saresti",
    "sarete",
    "sarà",
    "sarò",
    "se",
    "sei",
    "si",
    "sia",
    "siamo",
    "siano",
    "siate",
    "siete",
    "sono",
    "sta",
    "stai",
    "stando",
    "stanno",
    "starai",
    "staranno",
    "starebbe",
    "starebbero",
    "starei",
    "staremmo",
    "staremo",
    "stareste",
    "staresti",
    "starete",
    "starà",
    "starò",
    "stava",
    "stavamo",
    "stavano",
    "stavate",
    "stavi",
    "stavo",
    "stemmo",
    "stesse",
    "stessero",
    "stessi",
    "stessimo",
    "steste",
    "stesti",
    "stette",
    "stettero",
    "stetti",
    "stia",
    "stiamo",
    "stiano",
    "stiate",
    "sto",
    "su",
    "sua",
    "sue",
    "sugl",
    "sugli",
    "sui",
    "sul",
    "sull",
    "sulla",
    "sulle",
    "sullo",
    "suo",
    "suoi",
    "ti",
    "tra",
    "tu",
    "tua",
    "tue",
    "tuo",
    "tuoi",
    "tutti",
    "tutto",
    "un",
    "una",
    "uno",
    "vi",
    "voi",
    "vostra",
    "vostre",
    "vostri",
    "vostro",
    "è",
)
STOPWORDS_RUSSIAN = (
    "а",
    "без",
    "более",
    "больше",
    "будет",
    "будто",
    "бы",
    "был",
    "была",
    "были",
    "было",
    "быть",
    "в",
    "вам",
    "вас",
    "вдруг",
    "ведь",
    "во",
    "вот",
    "впрочем",
    "все",
    "всегда",
    "всего",
    "всех",
    "всю",
    "вы",
    "где",
    "да",
    "даже",
    "два",
    "для",
    "до",
    "другой",
    "его",
    "ее",
    "ей",
    "ему",
    "если",
    "есть",
    "еще",
    "ж",
    "же",
    "за",
    "зачем",
    "здесь",
    "и",
    "из",
    "или",
    "им",
    "иногда",
    "их",
    "к",
    "как",
    "какая",
    "какой",
    "когда",
    "конечно",
    "кто",
    "куда",
    "ли",
    "лучше",
    "между",
    "меня",
    "мне",
    "много",
    "может",
    "можно",
    "мой",
    "моя",
    "мы",
    "на",
    "над",
    "надо",
    "наконец",
    "нас",
    "не",
    "него",
    "нее",
    "ней",
    "нельзя",
    "нет",
    "ни",
    "нибудь",
    "никогда",
    "ним",
    "них",
    "ничего",
    "но",
    "ну",
    "о",
    "об",
    "один",
    "он",
    "она",
    "они",
    "опять",
    "от",
    "перед",
    "по",
    "под",
    "после",
    "потом",
    "потому",
    "почти",
    "при",
    "про",
    "раз",
    "разве",
    "с",
    "сам",
    "свою",
    "себе",
    "себя",
    "сейчас",
    "со",
    "совсем",
    "так",
    "такой",
    "там",
    "тебя",
    "тем",
    "теперь",
    "то",
    "тогда",
    "того",
    "тоже",
    "только",
    "том",
    "тот",
    "три",
    "тут",
    "ты",
    "у",
    "уж",
    "уже",
    "хорошо",
    "хоть",
    "чего",
    "чем",
    "через",
    "что",
    "чтоб",
    "чтобы",
    "чуть",
    "эти",
    "этого",
    "этой",
    "этом",
    "этот",
    "эту",
    "я",
)
STOPWORDS_SWEDISH = (
    "alla",
    "allt",
    "att",
    "av",
    "blev",
    "bli",
    "blir",
    "blivit",
    "de",
    "dem",
    "den",
    "denna",
    "deras",
    "dess",
    "dessa",
    "det",
    "detta",
    "dig",
    "din",
    "dina",
    "ditt",
    "du",
    "där",
    "då",
    "efter",
    "ej",
    "eller",
    "en",
    "er",
    "era",
    "ert",
    "ett",
    "från",
    "för",
    "ha",
    "hade",
    "han",
    "hans",
    "har",
    "henne",
    "hennes",
    "hon",
    "honom",
    "hur",
    "här",
    "i",
    "icke",
    "ingen",
    "inom",
    "inte",
    "jag",
    "ju",
    "kan",
    "kunde",
    "man",
    "med",
    "mellan",
    "men",
    "mig",
    "min",
    "mina",
    "mitt",
    "mot",
    "mycket",
    "ni",
    "nu",
    "när",
    "någon",
    "något",
    "några",
    "och",
    "om",
    "oss",
    "på",
    "samma",
    "sedan",
    "sig",
    "sin",
    "sina",
    "sitta",
    "själv",
    "skulle",
    "som",
    "så",
    "sådan",
    "sådana",
    "sådant",
    "till",
    "under",
    "upp",
    "ut",
    "utan",
    "vad",
    "var",
    "vara",
    "varför",
    "varit",
    "varje",
    "vars",
    "vart",
    "vem",
    "vi",
    "vid",
    "vilka",
    "vilkas",
    "vilken",
    "vilket",
    "vår",
    "våra",
    "vårt",
    "än",
    "är",
    "åt",
    "över",
)
STOPWORDS_NORWEGIAN = (
    "alle",
    "at",
    "av",
    "bare",
    "begge",
    "ble",
    "blei",
    "bli",
    "blir",
    "blitt",
    "både",
    "båe",
    "da",
    "de",
    "deg",
    "dei",
    "deim",
    "deira",
    "deires",
    "dem",
    "den",
    "denne",
    "der",
    "dere",
    "deres",
    "det",
    "dette",
    "di",
    "din",
    "disse",
    "ditt",
    "du",
    "dykk",
    "dykkar",
    "då",
    "eg",
    "ein",
    "eit",
    "eitt",
    "eller",
    "elles",
    "en",
    "enn",
    "er",
    "et",
    "ett",
    "etter",
    "for",
    "fordi",
    "fra",
    "før",
    "ha",
    "hadde",
    "han",
    "hans",
    "har",
    "hennar",
    "henne",
    "hennes",
    "her",
    "hjå",
    "ho",
    "hoe",
    "honom",
    "hoss",
    "hossen",
    "hun",
    "hva",
    "hvem",
    "hver",
    "hvilke",
    "hvilken",
    "hvis",
    "hvor",
    "hvordan",
    "hvorfor",
    "i",
    "ikke",
    "ikkje",
    "ingen",
    "ingi",
    "inkje",
    "inn",
    "inni",
    "ja",
    "jeg",
    "kan",
    "kom",
    "korleis",
    "korso",
    "kun",
    "kunne",
    "kva",
    "kvar",
    "kvarhelst",
    "kven",
    "kvi",
    "kvifor",
    "man",
    "mange",
    "me",
    "med",
    "medan",
    "meg",
    "meget",
    "mellom",
    "men",
    "mi",
    "min",
    "mine",
    "mitt",
    "mot",
    "mykje",
    "ned",
    "no",
    "noe",
    "noen",
    "noka",
    "noko",
    "nokon",
    "nokor",
    "nokre",
    "nå",
    "når",
    "og",
    "også",
    "om",
    "opp",
    "oss",
    "over",
    "på",
    "samme",
    "seg",
    "selv",
    "si",
    "sia",
    "sidan",
    "siden",
    "sin",
    "sine",
    "sitt",
    "sjøl",
    "skal",
    "skulle",
    "slik",
    "so",
    "som",
    "somme",
    "somt",
    "så",
    "sånn",
    "til",
    "um",
    "upp",
    "ut",
    "uten",
    "var",
    "vart",
    "varte",
    "ved",
    "vere",
    "verte",
    "vi",
    "vil",
    "ville",
    "vore",
    "vors",
    "vort",
    "vår",
    "være",
    "vært",
    "å",
)
STOPWORDS_CHINESE = (
    "一",
    "一下",
    "一些",
    "一切",
    "一则",
    "一天",
    "一定",
    "一方面",
    "一旦",
    "一时",
    "一来",
    "一样",
    "一次",
    "一片",
    "一直",
    "一致",
    "一般",
    "一起",
    "一边",
    "一面",
    "万一",
    "上下",
    "上升",
    "上去",
    "上来",
    "上述",
    "上面",
    "下列",
    "下去",
    "下来",
    "下面",
    "不一",
    "不久",
    "不仅",
    "不会",
    "不但",
    "不光",
    "不单",
    "不变",
    "不只",
    "不可",
    "不同",
    "不够",
    "不如",
    "不得",
    "不怕",
    "不惟",
    "不成",
    "不拘",
    "不敢",
    "不断",
    "不是",
    "不比",
    "不然",
    "不特",
    "不独",
    "不管",
    "不能",
    "不要",
    "不论",
    "不足",
    "不过",
    "不问",
    "与",
    "与其",
    "与否",
    "与此同时",
    "专门",
    "且",
    "两者",
    "严格",
    "严重",
    "个",
    "个人",
    "个别",
    "中小",
    "中间",
    "丰富",
    "临",
    "为",
    "为主",
    "为了",
    "为什么",
    "为什麽",
    "为何",
    "为着",
    "主张",
    "主要",
    "举行",
    "乃",
    "乃至",
    "么",
    "之",
    "之一",
    "之前",
    "之后",
    "之後",
    "之所以",
    "之类",
    "乌乎",
    "乎",
    "乘",
    "也",
    "也好",
    "也是",
    "也罢",
    "了",
    "了解",
    "争取",
    "于",
    "于是",
    "于是乎",
    "云云",
    "互相",
    "产生",
    "人们",
    "人家",
    "什么",
    "什么样",
    "什麽",
    "今后",
    "今天",
    "今年",
    "今後",
    "仍然",
    "从",
    "从事",
    "从而",
    "他",
    "他人",
    "他们",
    "他的",
    "代替",
    "以",
    "以上",
    "以下",
    "以为",
    "以便",
    "以免",
    "以前",
    "以及",
    "以后",
    "以外",
    "以後",
    "以来",
    "以至",
    "以至于",
    "以致",
    "们",
    "任",
    "任何",
    "任凭",
    "任务",
    "企图",
    "伟大",
    "似乎",
    "似的",
    "但",
    "但是",
    "何",
    "何况",
    "何处",
    "何时",
    "作为",
    "你",
    "你们",
    "你的",
    "使得",
    "使用",
    "例如",
    "依",
    "依照",
    "依靠",
    "促进",
    "保持",
    "俺",
    "俺们",
    "倘",
    "倘使",
    "倘或",
    "倘然",
    "倘若",
    "假使",
    "假如",
    "假若",
    "做到",
    "像",
    "允许",
    "充分",
    "先后",
    "先後",
    "先生",
    "全部",
    "全面",
    "兮",
    "共同",
    "关于",
    "其",
    "其一",
    "其中",
    "其二",
    "其他",
    "其余",
    "其它",
    "其实",
    "其次",
    "具体",
    "具体地说",
    "具体说来",
    "具有",
    "再者",
    "再说",
    "冒",
    "冲",
    "决定",
    "况且",
    "准备",
    "几",
    "几乎",
    "几时",
    "凭",
    "凭借",
    "出去",
    "出来",
    "出现",
    "分别",
    "则",
    "别",
    "别的",
    "别说",
    "到",
    "前后",
    "前者",
    "前进",
    "前面",
    "加之",
    "加以",
    "加入",
    "加强",
    "十分",
    "即",
    "即令",
    "即使",
    "即便",
    "即或",
    "即若",
    "却不",
    "原来",
    "又",
    "及",
    "及其",
    "及时",
    "及至",
    "双方",
    "反之",
    "反应",
    "反映",
    "反过来",
    "反过来说",
    "取得",
    "受到",
    "变成",
    "另",
    "另一方面",
    "另外",
    "只是",
    "只有",
    "只要",
    "只限",
    "叫",
    "叫做",
    "召开",
    "叮咚",
    "可",
    "可以",
    "可是",
    "可能",
    "可见",
    "各",
    "各个",
    "各人",
    "各位",
    "各地",
    "各种",
    "各级",
    "各自",
    "合理",
    "同",
    "同一",
    "同时",
    "同样",
    "后来",
    "后面",
    "向",
    "向着",
    "吓",
    "吗",
    "否则",
    "吧",
    "吧哒",
    "吱",
    "呀",
    "呃",
    "呕",
    "呗",
    "呜",
    "呜呼",
    "呢",
    "周围",
    "呵",
    "呸",
    "呼哧",
    "咋",
    "和",
    "咚",
    "咦",
    "咱",
    "咱们",
    "咳",
    "哇",
    "哈",
    "哈哈",
    "哉",
    "哎",
    "哎呀",
    "哎哟",
    "哗",
    "哟",
    "哦",
    "哩",
    "哪",
    "哪个",
    "哪些",
    "哪儿",
    "哪天",
    "哪年",
    "哪怕",
    "哪样",
    "哪边",
    "哪里",
    "哼",
    "哼唷",
    "唉",
    "啊",
    "啐",
    "啥",
    "啦",
    "啪达",
    "喂",
    "喏",
    "喔唷",
    "嗡嗡",
    "嗬",
    "嗯",
    "嗳",
    "嘎",
    "嘎登",
    "嘘",
    "嘛",
    "嘻",
    "嘿",
    "因",
    "因为",
    "因此",
    "因而",
    "固然",
    "在",
    "在下",
    "地",
    "坚决",
    "坚持",
    "基本",
    "处理",
    "复杂",
    "多",
    "多少",
    "多数",
    "多次",
    "大力",
    "大多数",
    "大大",
    "大家",
    "大批",
    "大约",
    "大量",
    "失去",
    "她",
    "她们",
    "她的",
    "好的",
    "好象",
    "如",
    "如上所述",
    "如下",
    "如何",
    "如其",
    "如果",
    "如此",
    "如若",
    "存在",
    "宁",
    "宁可",
    "宁愿",
    "宁肯",
    "它",
    "它们",
    "它们的",
    "它的",
    "安全",
    "完全",
    "完成",
    "实现",
    "实际",
    "宣布",
    "容易",
    "密切",
    "对",
    "对于",
    "对应",
    "将",
    "少数",
    "尔后",
    "尚且",
    "尤其",
    "就",
    "就是",
    "就是说",
    "尽",
    "尽管",
    "属于",
    "岂但",
    "左右",
    "巨大",
    "巩固",
    "己",
    "已经",
    "帮助",
    "常常",
    "并",
    "并不",
    "并不是",
    "并且",
    "并没有",
    "广大",
    "广泛",
    "应当",
    "应用",
    "应该",
    "开外",
    "开始",
    "开展",
    "引起",
    "强烈",
    "强调",
    "归",
    "当",
    "当前",
    "当时",
    "当然",
    "当着",
    "形成",
    "彻底",
    "彼",
    "彼此",
    "往",
    "往往",
    "待",
    "後来",
    "後面",
    "得",
    "得出",
    "得到",
    "心里",
    "必然",
    "必要",
    "必须",
    "怎",
    "怎么",
    "怎么办",
    "怎么样",
    "怎样",
    "怎麽",
    "总之",
    "总是",
    "总的来看",
    "总的来说",
    "总的说来",
    "总结",
    "总而言之",
    "恰恰相反",
    "您",
    "意思",
    "愿意",
    "慢说",
    "成为",
    "我",
    "我们",
    "我的",
    "或",
    "或是",
    "或者",
    "战斗",
    "所",
    "所以",
    "所有",
    "所谓",
    "打",
    "扩大",
    "把",
    "抑或",
    "拿",
    "按",
    "按照",
    "换句话说",
    "换言之",
    "据",
    "掌握",
    "接着",
    "接著",
    "故",
    "故此",
    "整个",
    "方便",
    "方面",
    "旁人",
    "无宁",
    "无法",
    "无论",
    "既",
    "既是",
    "既然",
    "时候",
    "明显",
    "明确",
    "是",
    "是不是",
    "是否",
    "是的",
    "显然",
    "显著",
    "普通",
    "普遍",
    "更加",
    "曾经",
    "替",
    "最后",
    "最大",
    "最好",
    "最後",
    "最近",
    "最高",
    "有",
    "有些",
    "有关",
    "有利",
    "有力",
    "有所",
    "有效",
    "有时",
    "有点",
    "有的",
    "有着",
    "有著",
    "望",
    "朝",
    "朝着",
    "本",
    "本着",
    "来",
    "来着",
    "极了",
    "构成",
    "果然",
    "果真",
    "某",
    "某个",
    "某些",
    "根据",
    "根本",
    "欢迎",
    "正在",
    "正如",
    "正常",
    "此",
    "此外",
    "此时",
    "此间",
    "毋宁",
    "每",
    "每个",
    "每天",
    "每年",
    "每当",
    "比",
    "比如",
    "比方",
    "比较",
    "毫不",
    "没有",
    "沿",
    "沿着",
    "注意",
    "深入",
    "清楚",
    "满足",
    "漫说",
    "焉",
    "然则",
    "然后",
    "然後",
    "然而",
    "照",
    "照着",
    "特别是",
    "特殊",
    "特点",
    "现代",
    "现在",
    "甚么",
    "甚而",
    "甚至",
    "用",
    "由",
    "由于",
    "由此可见",
    "的",
    "的话",
    "目前",
    "直到",
    "直接",
    "相似",
    "相信",
    "相反",
    "相同",
    "相对",
    "相对而言",
    "相应",
    "相当",
    "相等",
    "省得",
    "看出",
    "看到",
    "看来",
    "看看",
    "看见",
    "真是",
    "真正",
    "着",
    "着呢",
    "矣",
    "知道",
    "确定",
    "离",
    "积极",
    "移动",
    "突出",
    "突然",
    "立即",
    "第",
    "等",
    "等等",
    "管",
    "紧接着",
    "纵",
    "纵令",
    "纵使",
    "纵然",
    "练习",
    "组成",
    "经",
    "经常",
    "经过",
    "结合",
    "结果",
    "给",
    "绝对",
    "继续",
    "继而",
    "维持",
    "综上所述",
    "罢了",
    "考虑",
    "者",
    "而",
    "而且",
    "而况",
    "而外",
    "而已",
    "而是",
    "而言",
    "联系",
    "能",
    "能否",
    "能够",
    "腾",
    "自",
    "自个儿",
    "自从",
    "自各儿",
    "自家",
    "自己",
    "自身",
    "至",
    "至于",
    "良好",
    "若",
    "若是",
    "若非",
    "范围",
    "莫若",
    "获得",
    "虽",
    "虽则",
    "虽然",
    "虽说",
    "行为",
    "行动",
    "表明",
    "表示",
    "被",
    "要",
    "要不",
    "要不是",
    "要不然",
    "要么",
    "要是",
    "要求",
    "规定",
    "觉得",
    "认为",
    "认真",
    "认识",
    "让",
    "许多",
    "论",
    "设使",
    "设若",
    "该",
    "说明",
    "诸位",
    "谁",
    "谁知",
    "赶",
    "起",
    "起来",
    "起见",
    "趁",
    "趁着",
    "越是",
    "跟",
    "转动",
    "转变",
    "转贴",
    "较",
    "较之",
    "边",
    "达到",
    "迅速",
    "过",
    "过去",
    "过来",
    "运用",
    "还是",
    "还有",
    "这",
    "这个",
    "这么",
    "这么些",
    "这么样",
    "这么点儿",
    "这些",
    "这会儿",
    "这儿",
    "这就是说",
    "这时",
    "这样",
    "这点",
    "这种",
    "这边",
    "这里",
    "这麽",
    "进入",
    "进步",
    "进而",
    "进行",
    "连",
    "连同",
    "适应",
    "适当",
    "适用",
    "逐步",
    "逐渐",
    "通常",
    "通过",
    "造成",
    "遇到",
    "遭到",
    "避免",
    "那",
    "那个",
    "那么",
    "那么些",
    "那么样",
    "那些",
    "那会儿",
    "那儿",
    "那时",
    "那样",
    "那边",
    "那里",
    "那麽",
    "部分",
    "鄙人",
    "采取",
    "里面",
    "重大",
    "重新",
    "重要",
    "鉴于",
    "问题",
    "防止",
    "阿",
    "附近",
    "限制",
    "除",
    "除了",
    "除此之外",
    "除非",
    "随",
    "随着",
    "随著",
    "集中",
    "需要",
    "非但",
    "非常",
    "非徒",
    "靠",
    "顺",
    "顺着",
    "首先",
    "高兴",
)




import logging
import mmap
import os

import numpy as np

try:
    import ujson as json
except ImportError:
    import json

try:
    from tqdm.auto import tqdm
    TQDM_AVAILABLE = True
except ImportError:
    TQDM_AVAILABLE = False
    def tqdm(iterable=None, *args, **kwargs):
        return iterable


def change_extension(path, new_extension):
    path = str(path)
    return path.rpartition(".")[0] + new_extension


def find_newline_positions(path, show_progress=True, leave_progress=True):
    path = str(path)
    indexes = []
    with open(path, "r") as f:
        indexes.append(f.tell())
        pbar = tqdm(
            total=os.path.getsize(path),
            desc="Finding newlines for mmindex",
            unit="B",
            unit_scale=True,
            disable=not show_progress,
            leave=leave_progress,
        )

        while f.readline():
            t = f.tell()
            indexes.append(t)
        
            if pbar is not None:
                pbar.update(t - indexes[-2])
        
        if pbar is not None:
            pbar.close()

    return indexes[:-1]


def save_mmindex(indexes, path):
    path = str(path)
    index_file = change_extension(path, ".mmindex.json")
    with open(index_file, "w") as f:
        f.write(json.dumps(indexes))


def load_mmindex(path):
    path = str(path)
    index_file = change_extension(path, ".mmindex.json")
    with open(index_file, "r") as f:
        return json.loads(f.read())


# now we can jump to any line in the file thanks to the index and mmap
def get_line(
    path,
    index,
    mmindex,
    encoding="utf-8",
    file_obj=None,
    mmap_obj=None,
):
    path = str(path)
    if file_obj is None:
        file_obj = open(path, "r")
        CLOSE_FILE = True
    else:
        CLOSE_FILE = False

    if mmap_obj is None:
        mmap_obj = mmap.mmap(file_obj.fileno(), 0, access=mmap.ACCESS_READ)
        CLOSE_MMAP = True
    else:
        CLOSE_MMAP = False

    mmap_obj.seek(mmindex[index])
    result = mmap_obj.readline().decode(encoding)

    if CLOSE_MMAP:
        mmap_obj.close()

    if CLOSE_FILE:
        file_obj.close()

    return result


class JsonlCorpus:
    """
    A class to read a jsonl file line by line using mmap, allowing extremely fast
    access to any line in the file. For example, you could access the N-th line
    of a 10GB file in a fraction of a second, returning a dictionary.

    Example
    --------

    Traditioanally, you would read a jsonl file line by line like this:

    ```python
    import json
    data = [json.loads(line) for line in open("file.jsonl")]
    print(corpus[1000])
    ```

    This is memory inefficient and has a large overhead. Instead, you can use this class:

    ```python
    corpus = JsonlCorpus("file.jsonl")
    print(corpus[1000])

    ```

    Which only loads the line you need into memory, and is much faster.
    """

    def __init__(self, path, show_progress=True, leave_progress=True, save_index=True):
        self.path = path
        # if the index file does not exist, create it
        if os.path.exists(change_extension(path, ".mmindex.json")):
            self.mmindex = load_mmindex(path)
        else:
            logging.info("Creating index file for jsonl corpus")
            mmindex = find_newline_positions(
                path, show_progress=show_progress, leave_progress=leave_progress
            )
            if save_index:
                save_mmindex(mmindex, path)

            self.mmindex = mmindex

        self.file_obj = open(path, "r")
        self.mmap_obj = mmap.mmap(self.file_obj.fileno(), 0, access=mmap.ACCESS_READ)
        logging.info("Opened file and mmap objects")

    def __len__(self):
        return len(self.mmindex)

    def __getitem__(self, index):
        # handle multiple indices
        if isinstance(index, int):
            return json.loads(
                get_line(
                    self.path,
                    index,
                    self.mmindex,
                    file_obj=self.file_obj,
                    mmap_obj=self.mmap_obj,
                )
            )

        if isinstance(index, slice):
            return [self.__getitem__(i) for i in range(*index.indices(len(self)))]
        if isinstance(index, (list, tuple)):
            return [self.__getitem__(i) for i in index]
        if isinstance(index, np.ndarray):
            # if it's an ndarray, this means each element is an index, and the array can
            # be of any shape. thus, we should flatten it first, get the results as if it
            # was a list, and then reshape it back to the original shape
            index_flat = index.flatten().tolist()
            results = [self.__getitem__(i) for i in index_flat]
            reshaped = np.array(results).reshape(index.shape)
            return reshaped
        
        raise TypeError("Invalid index type")

    def __del__(self):
        if hasattr(self, "file_obj"):
            self.file_obj.close()
        if hasattr(self, "mmap_obj"):
            self.mmap_obj.close()
        logging.info("Closed file and mmap objects")


if __name__ == "__main__":
    # let's test the functions
    # random jsonl file
    file = "file.jsonl"
    # content is random uuids
    import uuid

    with open(file, "w") as f:
        for i in range(500):
            f.write(json.dumps({"uuid": str(uuid.uuid4())}) + "\n")

    # create the index
    mmindex = find_newline_positions(file)
    save_mmindex(mmindex, file)

    # read the first line
    # load the index
    mmindex = load_mmindex(file)
    print(get_line(file, 1, mmindex))

    print(get_line(file, 5, mmindex))



from setuptools import setup, find_packages

package_name = "bm25s"
version = {}
with open(f"{package_name}/version.py", encoding="utf8") as fp:
    exec(fp.read(), version)

with open("README.md", encoding="utf8") as fp:
    long_description = fp.read()

extras_require = {
    "core": ["jax[cpu]", "ujson", "tqdm", "PyStemmer"],
    "stem": ["PyStemmer"],
    "hf": ["huggingface_hub"],
    "dev": ["black"],
}
# Dynamically create the 'full' extra by combining all other extras
extras_require["full"] = sum(extras_require.values(), [])

setup(
    name=package_name,
    version=version["__version__"],
    author="Xing Han Lù",
    author_email=f"{package_name}@googlegroups.com",
    url=f"https://github.com/xhluca/{package_name}",
    description=f"An ultra-fast implementation of BM25 based on sparse matrices.",
    long_description=long_description,
    packages=find_packages(include=[f"{package_name}*"]),
    package_data={},
    install_requires=['scipy', 'numpy'],
    extras_require=extras_require,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    # Cast long description to markdown
    long_description_content_type="text/markdown",
)